#!/usr/bin/perl

######################################################################
package Cognition::uF;
######################################################################

use Cognition::MagicString;
use DateTime;
use DateTime::Format::ISO8601;
use DateTime::Format::Natural;
use DateTime::Format::Strptime;
use URI::Escape;
use XML::LibXML qw(:all);
use strict;

BEGIN
{
	our $VERSION = 0.1;
}


sub MAX_INCLUDE_ITERATIONS { 2; }


# Cleans away nested compound microformats. Any intentionally
# nested microformats (e.g. vcard class="agent vcard") should be
# dealt with BEFORE calling the destroyer! Because of the
# destructive nature of this function, make sure that you only
# use it on a clone of the real node.
sub destroyer
{
	my $element = shift;
	
	# Classes to be destroyed
	my @containers = qw(mfo vcard adr geo vcalendar vevent vtodo valarm
		vfreebusy hfeed hentry hslice hreview hresume xfolkentry);
	my %C;
	foreach my $c (@containers) { $C{$c}=1; }
	
	# Some classes may be retained, optionally.
	foreach my $c (@_)          { $C{$c}=0; }

	# Assemble them all into the regular expression of death.
	@containers = ();
	foreach my $c (keys %C) { push @containers, $c if $C{$c}; }
	my $regexp = join '|', @containers;
	$regexp = "\\b($regexp)\\b";
	
	# Destroy child elements matching the regular expression.
	foreach my $e ($element->getElementsByTagName('*'))
	{	
		next if ($e == $element);
		if ($e->getAttribute('class') =~ /$regexp/)
		{
			foreach my $c ($e->getElementsByTagName('*'))
			{
				$c->setAttribute('class', '');
				$c->setAttribute('rel', '');
			}
			my $newclass = $e->getAttribute('class');
			$newclass =~ s/$regexp//g;
			$e->setAttribute('class', $newclass);
		}
	}
}


sub data_patterns
{
	my $page = shift;
	my $root = shift;
	my $max_include_loops = shift || 2;

	# Expand microformat include pattern.
	my $incl_iterations = 0;
	my $replacements = 1;
	while (($incl_iterations < $max_include_loops) && $replacements)
	{
		$replacements = ufp_include($root, $page) + ufp_include2($root, $page);
		$incl_iterations++;
	}
	
	# Table cell headers pattern.
	Cognition::uF::ufp_table($root, $page);	

	# Magical data-X class pattern.
	Cognition::uF::ufp_data_class($root, $page);
}


sub ufp_data_class
{
	my $node = shift;
	my $page = shift;
	
	return
		unless ($page->{Profiles}->{'http://purl.org/uF/pattern-data-class/1'});
		
	foreach my $kid ($node->getElementsByTagName('*'))
	{
		my $classes = $kid->getAttribute('class');
		$classes =~ s/(^\s+|\s+$)//g;
		$classes =~ s/\s+/ /g;
		my @classes = split / /, $classes;
		map s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg, @classes;
		my @dataClasses = grep /^data\-/, @classes;
		next unless (@dataClasses);
		
		my $val = '';
		foreach my $d (@dataClasses)
		{
			$val = $d unless ((length $val) > (length $d));
		}
		
		$val =~ s/^data\-//;
		$kid->setAttribute('microformats-content', $val);
	}
}


sub ufp_table
{
	my $node = shift;
	my $page = shift;
	
	# Add node itself to list!
	my @elements = $node->getElementsByTagName('td');
	if ($node->tagName =~ /^t[dh]$/i)
		{ unshift @elements, $node; }
		
	foreach my $tag (@elements)
	{
		next unless (length $tag->getAttribute('headers'));
		
		my $headers = $tag->getAttribute('headers');
		$headers =~ s/(^\s+|\s+$)//g;
		$headers =~ s/\s+/ /g;
		my @headers = split / /, $headers;
		
		foreach my $H (@headers)
		{
			my $Htag = Cognition::HTMLParser::Utils::searchID($H, $page->{DOM});
			next unless ($Htag);
			next unless ($Htag->tagName =~ /^t[dh]$/i);
			
			my $new = $page->{DOM}->createElement('div');
			$new->setAttribute('class', $Htag->getAttribute('class'));
			foreach my $kid ($Htag->childNodes)
			{
				my $x = $kid->cloneNode(1);
				if ($kid->nodeType==XML_ELEMENT_NODE || $kid->nodeType==XML_TEXT_NODE)
				{
					my $r = $new->appendChild($x);
				}
			}
			$tag->appendChild($new);
		}
		
		$tag->setAttribute('headers', '');
	}
}

sub ufp_include
# Implements the standard microformats include pattern.
{
	my $node  = shift;
	my $page  = shift;
	my $class = shift || 'include';
	my $rv = 0;
	
	# For each link...
	my @links1 = $node->getElementsByTagName('a');
	my @links2 = $node->getElementsByTagName('object');
	my @links = (@links1, @links2);
	foreach my $link (@links)
	{	
		# Skip pattern if no class attribute found.
		my $classList = $link->getAttribute('class') || next;
		
		# We've found a use of the include pattern
		if ($classList =~ / (^|\s) $class (\s|$) /x)
		{
			my $href = length $link->getAttribute('href') ?
				$link->getAttribute('href') :
				$link->getAttribute('data') ;
				
			my $id = undef;
			if ($href =~ /^\#(.*)$/)
			{
				$id = $1;
			}
			else
			{
				next;
			}

			# find the included node
			my $replacement = Cognition::HTMLParser::Utils::searchID($id, $page->{DOM});
			next unless $replacement;

			# do not include it if it's an ancestor
			my $link_xpath = $link->getAttribute('_xpath');
			my $repl_xpath = $replacement->getAttribute('_xpath');
			next if (substr($link_xpath, 0, length $repl_xpath) eq $repl_xpath);

			# replace the including element with the included element
			$replacement = $replacement->cloneNode(1);
			$link->getParentNode->replaceChild($replacement, $link) && $rv++;
		}
	}
	
	# Return number of replacements made.
	return $rv;
	
} # /sub ufp_include


sub ufp_include2
# Implements the alternative microformats include pattern.
{
	my $node = shift;
	my $page = shift;
	my $classpfx = shift || '#';
	my $rv = 0;
	
	# Add node itself to list!
	my @elements = $node->getElementsByTagName('*');
	unshift @elements, $node;
	
	# For each element...
	foreach my $elem (@elements)
	{
		# Skip pattern if no class attribute found.
		my $classList = $elem->getAttribute('class');
		next unless ($classList =~ / $classpfx /x);

		my $atEnd = 0;
		
		$classList =~ s/(^\s|\s$)//g;
		$classList =~ s/\s+/ /g;
		my @classes = split / /, $classList;
		my @newClassList = ();
		
		foreach my $c (@classes)
		{
			if (substr($c,0,1) ne $classpfx && length($c)>=1)
			{
				push @newClassList, $c;
				$atEnd = 1;
				next;
			}
			
			my $id = $c; $id =~ s/^\#//x;
			my $replacement = Cognition::HTMLParser::Utils::searchID($id, $page->{DOM}) || next;

			# do not include it if it's an ancestor
			my $link_xpath = $elem->getAttribute('_xpath');
			my $repl_xpath = $replacement->getAttribute('_xpath');
			next if (substr($link_xpath, 0, length $repl_xpath) eq $repl_xpath);

			$replacement = $replacement->cloneNode(1);
			if ($atEnd)
			{
				$elem->appendChild($replacement) && $rv++;
			}
			else
			{
				$elem->insertBefore($replacement, $elem->getFirstChild) && $rv++;
			}
		}
		
		$elem->setAttribute('class', join(' ', @newClassList));
	}
	
	# Return number of replacements made.
	return $rv;
	
} # /sub ufp_include2


sub datetime
# Very tolerant DateTime parsing. Microformats are supposed to always use W3CDTF,
# but we'll be lenient and do our best to parse other ISO 8601 formats, and if
# that fails, even try to parse natural language.
{
	my $string = shift;
	my $tz     = shift || 'UTC';
	my $f;
	my $dt;

	$string = $string->{string} if (isms($string));
	
	$f = DateTime::Format::ISO8601->new;
	eval {
		my $isostring = $string;
		$isostring =~ s/([\+\-]\d\d)([014][50])$/\1\:\2/;	
		$dt = $f->parse_datetime($isostring);
	};

	if ($dt)
	{
		# Yay, we've parsed the datetime, but it helps to remember how
		# specific it was. First, let's normalise the format:
		$string =~ s/[^0-9DTtWw\:\-]//g;
		$string =~ s/^---/DDD/;
		$string =~ s/^--/DD/;
		$string =~ s/^-/D/;
		$string =~ s/([\+\-]\d\d\:?\d\d)$//;
		$string =~ s/[^0-9DTtWw]//g;
		$string = uc($string);
		
		if ($string =~ /^[1-9]\d{3}$/ || $string =~ /^D\d{2}$/)
			{ $dt->{resolution} = 'year'; }
		elsif ($string =~ /^[1-9]\d{5}$/ || $string =~ /^D?\d{4}$/ || $string =~ /^DD\d{2}$/)
			{ $dt->{resolution} = 'month'; }
		elsif ($string =~ /^\d{4}W\d{2}$/i || $string =~ /^\d{2}W\d{2}$/ || $string =~ /^D\d{1}W\d{2}$/ || $string =~ /^DW\d{2}$/)
			{ $dt->{resolution} = 'week'; }
		elsif ($string =~ /^\d{8}$/ || $string =~ /^\d{4}W\d{3}$/ || $string =~ /^\d{5,7}$/ || $string =~ /^D\d{3}$/ || $string =~ /^DW\d$/ || $string =~ /^D{2,3}\d{3,4}$/ || $string =~ /^D\d?W\d{3}$/ || $string =~ /^\d{2}W\d{3}$/)
			{ $dt->{resolution} = 'day'; }
		# hour, no minutes: not allowed by ISO8601, not parsed by perl formatter.
		elsif ($string =~ /T\d{4}$/)
			{ $dt->{resolution} = 'minute'; }
		elsif ($string =~ /T\d{6}$/)
			{ $dt->{resolution} = 'second'; }
		else
			{ $dt->{resolution} = 'nanosecond'; } # prob some false positives
	}

	else
	{
		$f = DateTime::Format::Natural->new(
			lang          => 'en',        # Should read this from source input
			prefer_future => 1,
			daytime       => {
										morning    =>  9,
										afternoon  => 13,
										evening    => 20
								  },
			time_zone     => "$tz"
		);
		$dt = $f->parse_datetime($string);
		$dt->{resolution} = 'second' unless ($dt->{resolution});
	}
	$dt = DateTime->now unless ($dt);
	$dt->{resolution} = 'nanosecond' unless ($dt->{resolution});
	
	my %pattern = (
	   year          => '%Y',
	   month         => '%Y-%m',
	   week          => '%F',
	   day           => '%F',
	   minute        => '%FT%H:%M%z',
	   second        => '%FT%H:%M:%S%z',
		nanosecond    => '%FT%T.%9N%z'
	);	
	$dt->set_formatter(DateTime::Format::Strptime->new(pattern=>$pattern{$dt->{resolution}}));
	
	return $dt;
}


1; #/package Cognition::HTMLParser::Microformats




